# Copyright 2015 Shannon Lucas
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

"""
Ordered data consists of a sequence of tokens that appear in a predictable
order.  Unlike fixed-field data, ordered data may contain tokens that are
optional or that may repeat a variable number of times.  These two
properties prevent a parser from relying on a token's positional index.  The
only guarantees provided by an ordered data grammar are the order that the
tokens will appear in, whether or not they are optional, and how many times a
specific token may repeat before a different token appears.

An example of ordered text is a U.S. street address::

    123 Broadway Ave., Apt. 15, Philadelphia, PA 19101

The elements of the address (street number, street name, city, etc.) appear in
a consistent order if the address is to be considered valid. The apartment
number may or may not be present in an address, but should appear in a
consistent position in the sequence in order to be valid.

A parser's grammar is defined by a sequence of ``GrammarElement`` instances.
Each of these elements is applied in the order it appears in the sequence.
Since the parser also extends ``AbstractCommand``, hierarchical ordered text
can be parsed by using one or more ordered parser instances as elements of
another parser's grammar.
"""

from collections.abc import Sequence

from ordtext.commands import AbstractCommand
from ordtext.exceptions import ParseCommandError, MinimumCountError


class OrderedParser(AbstractCommand):
    """Provides the engine for applying an ordered text grammar to a string of
    text data.
    """
    __slots__ = ("_grammar",)

    __REPR = "OrderedParser, {0} element grammar."

    def __init__(self, grammar):
        """Creates a new parser with the given grammar.

        :param Sequence[GrammarElement] grammar: a sequence of grammar elements
            in the order they should be applied to a set of ordered data.
        """
        self._grammar = grammar

    @property
    def grammar(self):
        """This parser's grammar.

        :rtype: tuple
        """
        return tuple(self._grammar)

    def parse(self, data, separator=None):
        """Parses a string of ordered text according to the parser's grammar.

        :param str data: the text to parse.
        :param string separator: the characters that separate the tokens within
            ``data``.  If no separator is given, whitespace will be used.

        :return: a tuple containing the parsed data (first element) and a
            sequence of any remaining tokens (second element).
        :rtype: (object, Sequence[str])
        """
        tokens = data.split(separator)
        return self.__call__(tokens)[0]

    def __repr__(self):
        """Returns the string representation of this instance.

        :rtype: str
        """
        return OrderedParser.__REPR.format(len(self.grammar))

    def __call__(self, tokens):
        """Applies the instance's grammar to a token sequence.

        :param Sequence[str] tokens: the sequence of tokens being parsed.

        :return: a tuple containing the parsed data (first element) and a
            sequence of any remaining tokens (second element).
        :rtype: (object, Sequence[str])
        """
        parsed = {}
        working = tokens

        for element in self.grammar:
            try:
                result, remainder = element(working)
            except ParseCommandError:
                result = None
                remainder = working
            except MinimumCountError as e:
                raise ParseCommandError(self.__class__.__name__, e)

            if element.retain:
                parsed[element.name] = result
            working = remainder

        return parsed, working
