# Copyright 2015 Shannon Lucas
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
import re

from ordtext.exceptions import ParseCommandError


class AbstractCommand(ABC):
    """Defines the interface the ordered text parser expects commands to
    expose.

    Instances of ``ParserCommand`` are invoked as functions; therefore
    sub-classes must implement the ``__call__`` method as defined in this
    class.

    Parser implementations must accept any function or object that conforms to
    the ``(Sequence) -> (object, Sequence)`` contract.  Parsers must not
    require that commands extend this class.
    """

    @abstractmethod
    def __call__(self, tokens):
        """Executes the command.

        Implementations consume one or more tokens from the head of the
        ``tokens`` sequence and then return the parsed data along with the
        unused tail of the sequence.

        :param Sequence[str] tokens: the sequence of tokens being parsed.
            Implementations should anticipate that this sequence may be empty
            depending on the parser engine implementation.

        :return: a tuple containing the parsed data (first element) and a
            sequence of any remaining tokens (second element).
        :rtype: (object, Sequence[str])

        :raises ParseCommandError: if the command could not parse the token
            at the beginning of the ``tokens`` sequence.

        .. note::
            Implementations should not modify the contents of the ``tokens``
            sequence itself.  Doing so may result in undefined behavior.
        """


class SimpleMatch(AbstractCommand):
    """Attempts to exactly match a token to a specific string literal.

    Commands of this type are useful in situations such as detecting section
    markers.
    """
    __slots__ = ("_match",)

    __REPR = "ExactMatch({0})"

    def __init__(self, *args):
        """Create an exact match command.

        :param str *args: the strings to look for when parsing tokens.
        """
        self._match = args

    @property
    def match(self):
        """The strings that this command matches on.

        :rtype: str
        """
        return self._match

    def __repr__(self):
        """Returns the string representation of this instance.

        :rtype: str
        """
        count = len(self.match)
        params = "{!r}, " * (count - 1) + "{!r}" if count > 1 else "{!r}"

        return SimpleMatch.__REPR.format(params.format(*self.match))

    def __call__(self, tokens):
        """Execute the command.

        :param Sequence[str] tokens: the sequence of tokens being parsed.

        :return: a tuple containing the matched token (first element) and a
            sequence of any remaining tokens (second element).
        :rtype: (str, Sequence[str])

        :raises ParseCommandError: if the first item in ``tokens`` does not
            exactly match the literal provided at instantiation.
        """
        if tokens[0] in self._match:
            return tokens[0], tokens[1:]
        else:
            raise ParseCommandError


class PatternMatch(AbstractCommand):
    """Attempts to match a token against a regular expression containing
    capture groups.

    The class may be used to capture numerically indexed groups or named
    groups, but not both simultaneously.
    """
    __slots__ = ("_pattern", "_by_name")

    __REPR = "PatternMatch(pattern={0!r}, by_name={1})"

    def __init__(self, pattern, by_name=False):
        """Create a pattern match command.

        :param pattern: the regular expression this instance should use to
            match tokens.
        :type pattern: str or re._pattern_type
        :param bool by_name: if set to ``True``, the instance will return
            a dictionary of named capture groups when called instead of a
            tuple of capture group values.
        """
        self._pattern = re.compile(pattern) if isinstance(pattern, str) \
            else pattern
        self._by_name = by_name

    @property
    def pattern(self):
        """The pattern (compiled) that this command matches on.

        :rtype: re._pattern_type
        """
        return self._pattern

    @property
    def by_name(self):
        """``True`` if the command produces a mapping of named capture group
        values.

        :rtype: bool
        """
        return self._by_name

    def __repr__(self):
        """Returns the string representation of this instance.

        :rtype: str
        """
        return PatternMatch.__REPR.format(self.pattern, self.by_name)

    def __call__(self, tokens):
        """Execute the command.

        :param Sequence[str] tokens: the sequence of tokens being parsed.

        :return: a tuple containing the match(es) from applying the command's
            pattern to the token (first element) and a sequence of any
            remaining tokens (second element).  The contents of the first
            element are determined according to these rules:

            - If ``by_name`` is set to ``True``, the first element will be a
              mapping of the named groups in the expression to their captured
              values.
            - If more than one capture group is present in the expression, the
              first element will be a tuple of the group matches.
            - If only one capture group is present in the expression, the first
              element will be the single value captured by that group.
            - If no capture groups are present, the first element will be the
              single value matched by the expression.
        :rtype: (str or Sequence[str] or Mapping[str, str], Sequence[str])

        :raises ParseCommandError: if the first item in ``tokens`` did not match
            the command's pattern.
        """
        match = self.pattern.fullmatch(tokens[0])

        if match:
            if self.by_name:
                result = match.groupdict()
            elif len(match.groups()) == 1:
                result = match.group(1)
            elif len(match.groups()):
                result = match.groups()
            else:
                result = match.group()
        else:
            raise ParseCommandError

        return result, tokens[1:]
