# Copyright 2015 Shannon Lucas
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

from collections.abc import Sequence

from ordtext.commands import AbstractCommand
from ordtext.exceptions import ParseCommandError, MinimumCountError


class GrammarElement(AbstractCommand):
    """Wraps a parsing command with contextual metadata such as the minimum
    and maximum sequential occurrences of the command's data.

    A sequence of these elements is used to create the grammar to drive an
    ordered text parser.
    """
    __slots__ = ("_command", "_name", "_min_count", "_max_count", "_retain",
                 "_description")

    __REPR = ("GrammarElement({command}, name={name!r}, "
              "min_count={min_count}, max_count={max_count}, "
              "retain={retain}, description={description!r}")

    def __init__(self, command, *, name="", min_count=0, max_count=None,
                 retain=True, description=""):
        """Constructs a new grammar element.

        :param command: the command this element wraps.
        :type command: AbstractCommand or FunctionType
        :param str name: a short, machine-readable name used by the parser to
            associate the data parsed by this command with a resulting data
            structure.
        :param int min_count: the number of times the wrapped command must
            appear in sequence.
        :param max_count: the maximum number of times this element is allowed
            to repeat.  If present, this value must be greater than or equal
            to ``min_value``.  A value of ``None`` indicates no limit.
        :type max_count: int or None
        :param bool retain: indicates that the parser should retain the result
            of this element's operations (default behavior).  If ``False``,
            parser implementations may choose not to include this result in
            the final parsed data.  This is useful for tokens that are
            syntactically required (such as begin and end tokens) but are not
            required in the parsed data.
        :param str description: an optional human-readable description for
            this grammar element.

        :raises AssertionError: if ``command`` is ``None``, if ``min_count``
            is less than zero or if ``max_count`` is less than ``min_count``.
        """
        assert command is not None
        assert min_count >= 0
        assert (max_count is None) or (max_count >= min_count)

        self._command = command
        self._name = name if name else command.__class__.__name__
        self._min_count = min_count
        self._max_count = max_count
        self._retain = retain
        self._description = description

    @property
    def command(self):
        """The command this element wraps.

        :rtype: AbstractCommand
        """
        return self._command

    @property
    def name(self):
        """A short, machine-readable name used by the parser to associate
        the data parsed by this command with a resulting data structure.

        :rtype: str
        """
        return self._name

    @property
    def min_count(self):
        """The number of times the wrapped command must appear in sequence.

        :rtype: int
        """
        return self._min_count

    @property
    def max_count(self):
        """The maximum number of times this element is allowed to repeat. A
        value of ``None`` indicates no limit.

        :rtype: int
        """
        return self._max_count

    @property
    def retain(self):
        """Indicates whether or not the parser should retain the result of
        this element's operations

        :rtype: bool
        """
        return self._retain

    @property
    def description(self):
        """An optional human-readable description for this grammar element.

        :rtype: str
        """
        return self._description

    @property
    def required(self):
        """Indicates that this element is required.

        :rtype: bool
        """
        return self.min_count > 0

    @property
    def singular(self):
        """Indicates that this element may occur only once.

        :rtype: bool
        """
        return self.max_count == 1

    def _process_one(self, tokens):
        """Processes a command one time.

        :param Sequence[str] tokens: the sequence of tokens being parsed.

        :return: a tuple containing the parsed data (first element) and a
            sequence of any remaining tokens (second element).
        :rtype: (object, Sequence[str])

        :raises ParseCommandError: if the command could not parse the token at
            the beginning of the ``tokens`` sequence.
        """
        try:
            return self.command(tokens)
        except ParseCommandError:
            raise
        except Exception as e:
            raise ParseCommandError(e)

    def _process_many(self, tokens):
        """Processes a command that may repeat.

        :param Sequence[str] tokens: the sequence of tokens being parsed.

        :return: a tuple containing the parsed data (first element) and a
            sequence of any remaining tokens (second element).
        :rtype: (object, Sequence[str])

        :raises ParseCommandError: if the command could not parse the token at
            the beginning of the ``tokens`` sequence.
        :raises MinimumCountError: if the wrapped command could not be
            successfully applied to the token sequence the expected number of
            times.
        """
        working = tokens
        parsed = []
        limit = self.max_count if self.max_count else len(tokens)

        for x in range(0, limit):
            try:
                result, remainder = self._process_one(working)
            except ParseCommandError:
                break
            else:
                parsed.append(result)
                working = remainder

        if len(parsed) < self.min_count:
            raise MinimumCountError(self.name, self.min_count, len(parsed))

        return parsed, working

    def __repr__(self):
        """Returns the string representation of this instance.

        :rtype: str
        """
        params = {
            "command": str(self.command),
            "name": self.name,
            "min_count": self.min_count,
            "max_count": self.max_count,
            "retain": self.retain,
            "description": self.description
        }
        return GrammarElement.__REPR.format(**params)

    def __call__(self, tokens):
        """Applies this grammar element to a token sequence.

        :param Sequence[str] tokens: the sequence of tokens being parsed.

        :return: a tuple containing the parsed data (first element) and a
            sequence of any remaining tokens (second element).  For elements
            that may repeat, the object returned in the first element will
            always be a ``Sequence`` of the types returned by the instance's
            command.  For elements that don't repeat, the returned object will
            be the type of object returned by the instance's command.
        :rtype: (object, Sequence[str])

        :raises ParseCommandError: if the command could not parse the token at
            the beginning of the ``tokens`` sequence.
        :raises MinimumCountError: if the wrapped command could not be
            successfully applied to the token sequence the expected number of
            times.
        """
        if self.singular:
            try:
                return self._process_one(tokens)
            except ParseCommandError:
                if self.required:
                    raise MinimumCountError(self.name, 1, 0)
                else:
                    raise
        else:
            return self._process_many(tokens)


def relax_grammar(grammar, *, relax_min=True, relax_max=False):
    """Creates a relaxed copy of a grammar.

    Grammars may be relaxed by removing the minimum and/or maximum count
    requirements for each element in the grammar.

    :param Sequence[GrammarElement] grammar: the grammar to relax.
    :param bool relax_min: ``True`` if the minimum count requirements for the
        grammar's elements should be relaxed (defaults to ``True``).
    :param bool relax_max: ``True`` if the maximum count limits for the
        grammar's elements should be relaxed (defaults to ``False``).

    :return: the relaxed grammar
    :rtype: Sequence[GrammarElement]
    """
    new_grammar = []

    for element in grammar:
        new_grammar.append(
            GrammarElement(
                element.command,
                name=element.name,
                min_count=0 if relax_min else element.min_count,
                max_count=None if relax_max else element.max_count,
                retain=element.retain,
                description=element.description
            )
        )

    return tuple(new_grammar)
